#include "esphome/core/color.h"

namespace esphome {

const Color Color::BLACK(0, 0, 0, 0);
const Color Color::WHITE(255, 255, 255, 255);

const Color COLOR_BLACK(0, 0, 0, 0);
const Color COLOR_WHITE(255, 255, 255, 255);

}  // namespace esphome
